/*
 * Copyright (c) 2006, TopCoder, Inc. All rights reserved.
 *
 * Author: TCSDEVELOPER, Ghostar
 * Version: 1.0
 */

using System;

namespace TopCoder.Document.Index.Persistence.DB
{
    /// <summary>
    /// This class extends the <see cref="DBPersistenceParameters"/> class to provide an extra <see cref="DocumentID"/>
    /// property as a supplement to the properties contained by its parent class.
    /// </summary>
    /// <remarks>
    /// This class is immutable and is thread safe.
    /// </remarks>
    public class DBDocumentIndexParameters : DBPersistenceParameters
    {
        /// <summary>
        /// This member variable holds the document ID of a specific document for the
        /// <see cref="DBDocumentIndexSource"/> instance to work against in the database. This value is set in the
        /// constructor and is accessed through the <see cref="DocumentID"/> property.
        /// </summary>
        private long documentId;

        /// <summary>
        /// The ID of the document in the database to load words for. It's stored here, as its use is directly related
        /// to creating an instance of the parameter class.
        /// </summary>
        internal const string DOCUMENTID_PROPERTY = "DocumentID";

        /// <summary>
        /// This is a getter only for the document ID of a specific document. This property is used by the
        /// <see cref="DBDocumentIndexSource"/> class to load a specific document's information from the database.
        /// </summary>
        public long DocumentID
        {
            get
            {
                return documentId;
            }
        }

        /// <summary>
        /// This constructor sets the properties of this class to the values of the parameters given.
        /// </summary>
        /// <param name="connectionName">Name of the connection to use.</param>
        /// <param name="documentId">The non-negative document ID to use.</param>
        /// <exception cref="ArgumentException">If <paramref name="connectionName"/> is empty.</exception>
        /// <exception cref="ArgumentOutOfRangeException">If <paramref name="documentId"/> is negative.</exception>
        /// <exception cref="ArgumentNullException">If <paramref name="connectionName"/> is null.</exception>
        public DBDocumentIndexParameters(string connectionName, long documentId)
            : base(connectionName)
        {
            if (documentId < 0)
            {
                throw new ArgumentOutOfRangeException("documentId", documentId, "Negative values not allowed.");
            }
            this.documentId = documentId;
        }

        /// <summary>
        /// Returns a hash code, based on the connection name and document id.
        /// </summary>
        /// <returns>A hash code.</returns>
        public override int GetHashCode()
        {
            return base.GetHashCode() ^ documentId.GetHashCode();
        }

        /// <summary>
        /// This method overloads the standard <see cref="Equals"/> method to see if two
        /// <see cref="DBDocumentIndexParameters"/> instances are the same, based on the connection name and document
        /// ID.
        /// </summary>
        /// <param name="obj">The object to compare against.</param>
        /// <returns>True if <paramref name="obj"/> is equal to this object, false if not.</returns>
        public override bool Equals(object obj)
        {
            DBDocumentIndexParameters objCasted = obj as DBDocumentIndexParameters;
            if (objCasted == null)
            {
                return false;
            }

            return base.Equals(objCasted) && this.DocumentID == objCasted.DocumentID;
        }
    }
}