/*
 * Copyright (c) 2006, TopCoder, Inc. All rights reserved.
 *
 * Author: TCSDEVELOPER, Ghostar
 * Version: 1.0
 */

using System;

namespace TopCoder.Document.Index.Persistence.DB
{
    /// <summary>
    /// This class holds options used to initialize both <see cref="DBDocumentIndexSource"/> and
    /// <see cref="DBCollectionIndexSource"/>. The connection name property is used to get a connection from the
    /// Connection Factory component.
    /// </summary>
    /// <remarks>
    /// This class is immutable as its member variable can't change after initialization.
    /// </remarks>
    public class DBPersistenceParameters : IParameters
    {
        /// <summary>
        /// This member variable holds the connection name, used to create a connection from the Connection Factory.
        /// This value is set in the constructor and accessed by the user using the <see cref="ConnectionName"/>
        /// property.
        /// </summary>
        private string connectionName;

        /// <summary>
        /// The name of the connection to get from the Connection Factory is stored in this configuration manager
        /// property. It's stored here, as its use is directly related to creating an instance of the parameter class.
        /// </summary>
        internal const string CONNECTIONNAME_PROPERTY = "DBConnectionName";

        /// <summary>
        /// This property accesses the <see cref="connectionName"/> member variable. It is used by the DB persistence
        /// classes as the connection name to use when getting a connection from the Connection Factory component.
        /// </summary>
        /// <value>
        /// The connection name to use.
        /// </value>
        public string ConnectionName
        {
            get
            {
                return connectionName;
            }
        }

        /// <summary>
        /// This constructor just initializes the values of the member variable to the value of the parameter provided.
        /// </summary>
        /// <param name="connectionName">Name of the connection to use.</param>
        /// <exception cref="ArgumentException">If <paramref name="connectionName"/> is empty.</exception>
        /// <exception cref="ArgumentNullException">If <paramref name="connectionName"/> is null.</exception>
        public DBPersistenceParameters(string connectionName)
        {
            if (connectionName == null)
            {
                throw new ArgumentNullException("connectionName");
            }
            if (connectionName.Trim().Length == 0)
            {
                throw new ArgumentException("Empty string specified.", "connectionName");
            }

            this.connectionName = connectionName;
        }

        /// <summary>
        /// Returns a hash code, based on the connection name.
        /// </summary>
        /// <returns>A hash code.</returns>
        public override int GetHashCode()
        {
            return connectionName.GetHashCode();
        }

        /// <summary>
        /// This method overloads the standard <see cref="Equals"/> method to see if two
        /// <see cref="DBPersistenceParameters"/> instances are the same, based on the connection name.
        /// </summary>
        /// <param name="obj">The object to compare against.</param>
        /// <returns>True if <paramref name="obj"/> is equal to this object, false if not.</returns>
        public override bool Equals(object obj)
        {
            DBPersistenceParameters objCasted = obj as DBPersistenceParameters;
            if (objCasted == null)
            {
                return false;
            }

            // "is" won't work, because then a DBPersistenceParameters instance would be equal to
            // DBDocumentIndexParameters with the same connection name.
            if (!this.GetType().Equals(objCasted.GetType()))
            {
                return false;
            }

            return this.ConnectionName == objCasted.ConnectionName;
        }
    }
}