/*
 * Copyright (c) 2006, TopCoder, Inc. All rights reserved.
 *
 * Author: TCSDEVELOPER, Ghostar
 * Version: 1.0
 */

using System;
using System.Collections;
using System.Globalization;
using System.IO;
using System.Text;
using System.Xml;
using TopCoder.Document.Index.WordStreams;

namespace TopCoder.Document.Index.Persistence.Xml
{
    /// <summary>
    /// This class is an XML-based implementation of <see cref="IDocumentIndexSource"/>. It saves to, and loads from,
    /// one or more XML files, depending on a limit defined by the user.
    /// </summary>
    /// <remarks>
    /// Besides just the standard saving to an XML file, this class also provides the option to split the saving of the
    /// document index to multiple XML files, all linked from the main file. This allows us a way to get around file
    /// system size limits on a single file. This class is thread safe, as the access to the words dictionary is
    /// synchronized.
    /// </remarks>
    public class XmlDocumentIndexSource : IDocumentIndexSource, IComparer
    {
        /// <summary>
        /// The Uri of the namespace for all XML files.
        /// </summary>
        private const string XML_NAMESPACE = "http://www.topcoder.com";

        /// <summary>
        /// The root of the main XML file for a document index.
        /// </summary>
        private const string DOCUMENT_ELEMENT = "Document";

        /// <summary>
        /// The element identifying the compare options to use.
        /// </summary>
        private const string COMPAREOPTIONS_ELEMENT = "CompareOptions";

        /// <summary>
        /// The element identifying the culture to use for string comparisons.
        /// </summary>
        private const string CULTURE_ELEMENT = "Culture";

        /// <summary>
        /// The main element of a main file. The words themselves are in the word list, in an inner element.
        /// </summary>
        private const string WORDS_ELEMENT = "Words";

        /// <summary>
        /// The inner element containing the list of Word and Locations elements.
        /// </summary>
        private const string WORDSLIST_ELEMENT = "WordsList";

        /// <summary>
        /// The main element of the list of subfiles in a main file.
        /// </summary>
        private const string FILES_ELEMENT = "Files";

        /// <summary>
        /// The content of this element is the file path to a specific file, specified as a path relative to that
        /// of the main file.
        /// </summary>
        private const string FILEPATH_ELEMENT = "FilePath";

        /// <summary>
        /// The template for how to construct the name of a subfile while saving. This is rather simple, the main file
        /// name/path as {0} and then {1} being a (random) GUID. <see cref="String.Format"/> is used to "parse" this
        /// template.
        /// </summary>
        private const string SUBFILENAME_TEMPLATE = "{0}{1}.xml";

        /// <summary>
        /// In a subfile, the Word/Locations list is placed within a "WordsList" element. Hence, this element is
        /// equivalent to the WordsList element in a main file, not a Words list.
        /// </summary>
        private const string SUBFILEWORDS_ELEMENT = "WordsList";

        /// <summary>
        /// The element specifying the verbatim form of a word.
        /// </summary>
        private const string WORD_ELEMENT = "Word";

        /// <summary>
        /// The element containing a list of integer word positions.
        /// </summary>
        private const string LOCATIONS_ELEMENT = "Locations";

        /// <summary>
        /// The element containing a single integer position.
        /// </summary>
        private const string LOCATION_ELEMENT = "Location";

        /// <summary>
        /// This property returns a newly created <see cref="IndexInfo"/> instance, created with the assembly and name
        /// of this class, along with a <see cref="SimpleIndexParameters"/> instance that holds the path to the file
        /// that the current instance is currently working against.
        /// </summary>
        /// <exception cref="InvalidOperationException">If no file path has been set for this instance, and hence it
        /// cannot be recreated by the framework.</exception>
        /// <value>
        /// An <see cref="IndexInfo"/> instance, populated with relevant data.
        /// </value>
        public IndexInfo IndexInfo
        {
            get
            {
                if (FilePath == null)
                {
                    throw new InvalidOperationException("Saving impossible if no file path has been specified.");
                }

                // Just create an instance. Note the fully qualified assembly name.
                return new IndexInfo(GetType().Assembly.FullName, GetType().FullName,
                    new SimpleIndexParameters(FilePath));
            }
        }

        /// <summary>
        /// This property provides read and write access to the culture member variable of this class, used to set the
        /// culture to use for string comparisons.
        /// </summary>
        /// <exception cref="ArgumentNullException">If <paramref name="value"/> is null.</exception>
        /// <exception cref="InvalidOperationException">If the options are changed for a non-empty index. This is
        /// dangerous, since changes of sorting order could invalidate existing data.</exception>
        /// <value>
        /// The default value, if nothing else is specified, is en-US.
        /// </value>
        public CultureInfo Culture
        {
            get
            {
                return culture;
            }
            set
            {
                if (value == null)
                {
                    throw new ArgumentNullException("value");
                }
                lock (SyncRoot)
                {
                    if (words.Count > 0)
                    {
                        throw new
                            InvalidOperationException("Sorting order can only be changed when no words are present.");
                    }
                    culture = value;
                }
            }
        }

        /// <summary>
        /// This property holds compare options.  It controls, together with the <see cref="Culture"/> property, how
        /// string comparisons are performed.
        /// </summary>
        /// <exception cref="InvalidOperationException">If the options are changed for a non-empty index. This is
        /// dangerous, since changes of sorting order could invalidate existing data.</exception>
        /// <value>
        /// The default value is <see cref="System.Globalization.CompareOptions.None"/>, but any value can be used.
        /// </value>
        public CompareOptions CompareOptions
        {
            get
            {
                return compareOptions;
            }
            set
            {
                lock (SyncRoot)
                {
                    if (words.Count > 0)
                    {
                        throw new
                            InvalidOperationException("Sorting order can only be changed when no words are present.");
                    }
                    compareOptions = value;
                }
            }
        }

        /// <summary>
        /// This private member variable holds the culture to use when comparing two strings. This allows us to
        /// provide support for multiple languages and alphabets when searching through a given document. This member
        /// variable is initialized to a default value, and it can be changed through the <see cref="Culture"/>
        /// property.
        /// </summary>
        private CultureInfo culture = new CultureInfo("en-US");

        /// <summary>
        /// This member variable holds options, beyond the culture, for comparing strings in this particular class.
        /// This value is initialized to a default, and it can be accessed using the <see cref="CompareOptions"/>
        /// property.
        /// </summary>
        private CompareOptions compareOptions = CompareOptions.None;

        /// <summary>
        /// This member variable holds the file path to where the index should be saved. This member variable is set in
        /// the constructor, as well as using the <see cref="FilePath"/> property. It is accessed by the
        /// <see cref="Save"/> method when saving to a particular file.
        /// </summary>
        private string filePath = null;

        /// <summary>
        /// This member variable holds the number of sub files to split the words across when saving them to the file
        /// system. This allows the user to get as fine grained as they want, in terms of individual file size. This
        /// value defaults to 0, in which case all the words will be in the original file. This value can be retrieved
        /// and set using the <see cref="NumberOfSubFiles"/> property.
        /// </summary>
        private int numberOfSubFiles = 0;

        /// <summary>
        /// This property is used by the user to set how many sub files the save mechanism should break the words
        /// across. This value has to be at least zero, where zero indicates that no subfiles are used and all words
        /// are in the XML file created. This value also has to be less than or equal to the number of words in the
        /// document.
        /// </summary>
        /// <exception cref="ArgumentOutOfRangeException">If <paramref name="value"/> is less than 0 or greater than
        /// the number of words.</exception>
        /// <value>
        /// 0, to indicate no subfiles, or [1..number of words] to indicate the number of subfiles.
        /// </value>
        public int NumberOfSubFiles
        {
            get
            {
                return numberOfSubFiles;
            }
            set
            {
                // Lock, as the number of sub files is used while saving and loading
                lock (SyncRoot)
                {
                    if (value < 0 || value > words.Count)
                    {
                        throw new ArgumentOutOfRangeException("value", value,
                            "Valid values are >= 0 and <= the number of words.");
                    }
                }
            }
        }

        /// <summary>
        /// This property provides access to the internally stored file path, used to set the location where the XML
        /// will be saved.
        /// </summary>
        /// <exception cref="ArgumentException">If <paramref name="value"/> is empty.</exception>
        /// <exception cref="ArgumentException">If <paramref name="value"/> is an invalid path, for example pointing
        /// to a directory.</exception>
        /// <exception cref="ArgumentNullException">If <paramref name="value"/> is null.</exception>
        /// <value>
        /// A non-null, non-empty path.
        /// </value>
        public string FilePath
        {
            get
            {
                return filePath;
            }
            set
            {
                if (value == null)
                {
                    throw new ArgumentNullException("value");
                }
                value = value.Trim();
                if (value.Length == 0)
                {
                    throw new ArgumentException("Empty string specified.", value);
                }

                try
                {
                    DirectoryInfo dirInfo = new DirectoryInfo(value);
                    if (dirInfo.Exists)
                    {
                        throw new ArgumentException("Path points to a directory.");
                    }
                }
                catch (ArgumentException)
                {
                    throw;
                }
                catch (Exception e)
                {
                    throw new ArgumentException("Path invalid or access denied.", "value", e);
                }

                // Don't possible do anything with the current path while loading/saving
                lock (SyncRoot)
                {
                    if (value != filePath)
                    {
                        // Former subfiles not relevant if we change the file name.
                        subFiles.Clear();
                    }
                    filePath = value;
                }
            }
        }

        /// <summary>
        /// This member variable holds the loaded words, loaded from the XML file, that this instance is working
        /// against. The keys in this dictionary are the words, and the values are <see cref="IList"/> instances of
        /// integer locations in the document where this word can be found. This collection is loaded by the
        /// <see cref="Load"/> method and accessed by the indexers in this class.
        /// </summary>
        private SortedList words;

        /// <summary>
        /// A list of the subfiles used for this filename. Used for proper replacing when saving.
        /// </summary>
        private IList subFiles = new ArrayList();

        /// <summary>
        /// This property provides the object to use when synchronizing access to the word collection.
        /// </summary>
        /// <value>
        /// A reference to the sync root on the internal collection.
        /// </value>
        public object SyncRoot
        {
            get
            {
                return words.SyncRoot;
            }
        }

        /// <summary>
        /// This constructor initializes the class using the <paramref name="filePath"/> given. That property is the
        /// path where the XML data is saved to. No loading is performed in the constructor.
        /// </summary>
        /// <param name="filePath">The filepath to use when loading.</param>
        /// <exception cref="ArgumentException">If <paramref name="filePath"/> is empty.</exception>
        /// <exception cref="ArgumentException">If <paramref name="filePath"/> is an invalid path, for example pointing
        /// to a directory.</exception>
        /// <exception cref="ArgumentNullException">If <paramref name="filePath"/> is null.</exception>
        public XmlDocumentIndexSource(string filePath) : this()
        {
            FilePath = filePath;
        }

        /// <summary>
        /// Default constructor, intended to be used with <see cref="Load"/> to load a specific file afterwards.
        /// </summary>
        public XmlDocumentIndexSource()
        {
            words = SortedList.Synchronized(new SortedList(this));
        }

        /// <summary>
        /// This method adds the given word to the XML document. This method should just update the internally stored
        /// words list, creating a new <see cref="ArrayList"/>.
        /// </summary>
        /// <param name="wordInfo">The word info to add.</param>
        /// <exception cref="ArgumentNullException">If <paramref name="wordInfo"/> is null.</exception>
        /// <exception cref="ArgumentException">If the word in <paramref name="wordInfo"/> is empty.</exception>
        public void Add(WordInfo wordInfo)
        {
            if (wordInfo == null)
            {
                throw new ArgumentNullException("wordInfo");
            }
            string word = wordInfo.Word.Trim();
            if (word.Length == 0)
            {
                throw new ArgumentException("Empty word string specified.", "wordInfo");
            }

            lock (SyncRoot)
            {
                IList list = words[word] as IList;

                if (list == null)
                {
                    // Create a new list. Let's assume that it's very frequent that words only occur once (that's quite
                    // true for most data in real corpora), then it will grow just like normal.
                    list = new ArrayList(1);
                    words[word] = list;

                    // Now, we can add the new position to the list, no matter how it got there.
                }

                list.Add(wordInfo.Position);
            }
        }

        /// <summary>
        /// This method removes the given word and its locations from the words collection.
        /// </summary>
        /// <param name="word">The word to remove.</param>
        /// <exception cref="ArgumentNullException">If <paramref name="word"/> is null.</exception>
        /// <exception cref="ArgumentException">If <paramref name="word"/> is an empty string.</exception>
        public void Remove(string word)
        {
            if (word == null)
            {
                throw new ArgumentNullException("word");
            }
            word = word.Trim();
            if (word.Length == 0)
            {
                throw new ArgumentException("Empty string specified.", "word");
            }

            // The words collection is synchronizing itself, common to our SyncRoot
            words.Remove(word);
        }

        /// <summary>
        /// This method returns whether or not the document represented by this class contains the given word.
        /// </summary>
        /// <param name="word">The word to check for.</param>
        /// <exception cref="ArgumentNullException">If <paramref name="word"/> is null.</exception>
        /// <exception cref="ArgumentException">If <paramref name="word"/> is an empty string.</exception>
        /// <returns>True if it exists, false if not.</returns>
        public bool Contains(string word)
        {
            if (word == null)
            {
                throw new ArgumentNullException("word");
            }
            word = word.Trim();
            if (word.Length == 0)
            {
                throw new ArgumentException("Empty string specified.", "word");
            }

            // Lock access to the internal representation, the list itself might change after retrieval otherwise
            lock (SyncRoot)
            {
                IList list = words[word] as IList;

                // It doesn't exist
                if (list == null)
                {
                    return false;
                }

                // If there is such a list, and it's non-empty, return true.
                return list.Count > 0;
            }
        }

        /// <summary>
        /// This method clears all words from the document.
        /// </summary>
        public void Clear()
        {
            // Synchronization is handled by the internal object
            words.Clear();
        }

        /// <summary>
        /// This method just returns the enumerator for the words collection, <see cref="IDictionary.GetEnumerator"/>.
        /// This method can be used by the user to enumerate through all words in a given document index.
        /// </summary>
        /// <remarks>
        /// Please note that using an enumerator is always inherently thread-unsafe, if any
        /// editing is performed on the underlying structure. Lock on <c>SyncRoot</c> while consuming the enumerator,
        /// and avoid making any changes from the current thread, if needed.
        /// </remarks>
        /// <returns>
        /// An enumerator instance.
        /// </returns>
        public IDictionaryEnumerator GetEnumerator()
        {
            // We can't synchronize this any better than what's performed by the collection
            return words.GetEnumerator();
        }

        /// <summary>
        /// Reads until an element start, or a file end, is found.
        /// </summary>
        /// <param name="reader">The reader to read from.</param>
        private static void ReadToStartElement(XmlReader reader)
        {
            while (reader.NodeType != XmlNodeType.Element && reader.Read())
            {
            }
        }

        /// <summary>
        /// This method loads the data from the persistent store, in this case an XML file. The parameter given should
        /// be a <see cref="SimpleIndexParameters"/> instance that contains a file path that is used to initialize the
        /// file path property. The list of words in this file, and in any "linked" files for a spanned XML structure,
        /// needs to be added.
        /// </summary>
        /// <remarks>
        /// <para>
        /// If the files are basically stored in monotonically increasing order, the total loading time should be
        /// O(N log N). This is quite acceptable. If the files are unordered, the loading performance may be O(N^2).
        /// The O(N^2) could be eliminated if we accepted doing an additional copy. Doing that in the general case
        /// would however possibly be quite expensive, especially when handling huge data sets. Any files saved by this
        /// component will be sorted in a favorable way.
        /// </para>
        /// <para>
        /// Please note that the state will be undefined if the load operation fails.
        /// </para>
        /// </remarks>
        /// <exception cref="ArgumentNullException">If <paramref name="parameters"/> is null.</exception>
        /// <exception cref="ArgumentException">If <paramref name="parameters"/> is not a
        /// <see cref="SimpleIndexParameters"/> instance, or the path is invalid, or the file doesn't exist.
        /// </exception>
        /// <exception cref="LoadIndexDataException">Any error while parsing XML.</exception>
        /// <param name="parameters">The parameters specifying the path to load from.</param>
        public void Load(IParameters parameters)
        {
            if (parameters == null)
            {
                throw new ArgumentNullException("parameters");
            }

            SimpleIndexParameters pathParams = parameters as SimpleIndexParameters;
            if (pathParams == null)
            {
                throw new ArgumentException("Specified parameters not of SimpleIndexParameters type.", "parameters");
            }

            lock (SyncRoot)
            {
                // Set the new path and get the directory part of it to construct the full paths to the subfiles
                string relativePath = null;

                try
                {
                    if (!File.Exists(pathParams.FileName))
                    {
                        throw new ArgumentException("The specified file doesn't exist.", "parameters");
                    }
                    relativePath = Path.GetDirectoryName(pathParams.FileName);
                }
                catch (ArgumentException)
                {
                    throw;
                }
                catch (Exception e)
                {
                    throw new ArgumentException("Error accessing file.", "parameters", e);
                }

                FilePath = pathParams.FileName;

                // Set newSubFiles to subFiles when all reading has been done
                IList newSubFiles = new ArrayList();

                // This is the safest solution, to avoid ever deleting the incorrect files while saving
                subFiles.Clear();
                Clear();
                XmlReader reader = null;

                // Now, we're ready to read
                try
                {
                    reader = new XmlTextReader(filePath);

                    bool readRoot = false;

                    // An XmlReader is used to avoid loading a complete DOM in memory (XPath and other methods not
                    // supported directly in .NET before 2.0)
                    while (reader.Read())
                    {
                        // Most of the action takes place when the root element has been identified
                        if (reader.NodeType == XmlNodeType.Element)
                        {
                            // Document
                            reader.ReadStartElement(DOCUMENT_ELEMENT, XML_NAMESPACE);

                            if (readRoot)
                            {
                                throw new XmlException(String.Format("Double root elements in {0}.", filePath));
                            }
                            readRoot = true;

                            ReadToStartElement(reader);
                            // CompareOptions, trimming not needed as Enum.Parse does that.
                            CompareOptions = (CompareOptions) Enum.Parse(typeof(CompareOptions),
                                reader.ReadElementString(COMPAREOPTIONS_ELEMENT, XML_NAMESPACE));

                            ReadToStartElement(reader);
                            // Culture
                            Culture = new CultureInfo(
                                reader.ReadElementString(CULTURE_ELEMENT, XML_NAMESPACE).Trim());

                            ReadToStartElement(reader);
                            // Words
                            reader.ReadStartElement(WORDS_ELEMENT, XML_NAMESPACE);

                            ReadToStartElement(reader);
                            // WordsList
                            // This element is optional, the actual list of words
                            if (reader.LocalName == WORDSLIST_ELEMENT && reader.NamespaceURI == XML_NAMESPACE)
                            {
                                reader.Read();
                                GetWordsFromElement(reader);
                            }

                            // This element is also optional, the list of inner files
                            ReadToStartElement(reader);

                            // Reset the number of sub files, soon we will know the real number in this file.
                            numberOfSubFiles = 0;
                            if (reader.LocalName == FILES_ELEMENT && reader.NamespaceURI == XML_NAMESPACE)
                            {
                                reader.Read();
                                while (reader.NodeType != XmlNodeType.EndElement)
                                {
                                    if (reader.NodeType == XmlNodeType.Element)
                                    {
                                        numberOfSubFiles++;

                                        // Paths are relative, reconstruct a full path
                                        string path =
                                            reader.ReadElementString(FILEPATH_ELEMENT, XML_NAMESPACE).Trim();

                                        path = Path.Combine(relativePath, path);
                                        newSubFiles.Add(path);
                                        GetWordsFromFile(path);
                                    }
                                    else
                                    {
                                        reader.Read();
                                    }
                                }
                            }
                        }
                    }
                    if (!readRoot)
                    {
                        throw new XmlException(String.Format("No root element found in {0}.", filePath));
                    }

                }
                catch (Exception e)
                {
                    // Avoid the risk of saving in an undefined state. The previous state is not restored (this could
                    // be prohibitively expensive with large datasets).
                    filePath = null;
                    subFiles.Clear();
                    words.Clear();

                    throw new LoadIndexDataException(parameters, e, "Error loading from XML.");
                }
                finally
                {
                    if (reader != null)
                    {
                        reader.Close();
                    }
                }

                subFiles = newSubFiles;
            }
        }

        /// <summary>
        /// This method saves out the internally stored <see cref="IDictionary"/> as one or several XML files,
        /// following the Document schema.
        /// </summary>
        /// <remarks>
        /// If <see cref="NumberOfSubFiles"/> is greater than 0, the words are split into that many groups, with each
        /// group going to a different file. The path for each sub file should be the name of the file specified in the
        /// <see cref="FilePath"/> member variable, appended with a GUID, with an extension of "XML". The split is done
        /// to put approximately the same number of words in each file.
        /// </remarks>
        /// <exception cref="SaveIndexDataException">Any error while saving.</exception>
        /// <exception cref="InvalidOperationException">If saving is called in a state where no file name is present.
        /// </exception>
        public void Save()
        {
            if (FilePath == null)
            {
                throw new InvalidOperationException("Saving impossible if no file path has been specified.");
            }

            try
            {
                lock (SyncRoot)
                {
                    // These are stored WITH paths
                    IList newSubFiles = new ArrayList();

                    // These are just the Guid part of the file names, used to construct the relative paths.
                    IList guids = new ArrayList();

                    // Limit the number of sub files when actually writing.
                    if (words.Count < numberOfSubFiles)
                    {
                        numberOfSubFiles = words.Count;
                    }

                    // Write the sub-files
                    if (NumberOfSubFiles > 0)
                    {
                        // We write countPerFile words per file, but for the countOfExtras first, we write one more.
                        // This distributes the words as evenly as possible.

                        int countPerFile = words.Count / NumberOfSubFiles;
                        int countOfExtras = words.Count % NumberOfSubFiles;
                        int nowAt = 0;

                        // Strip away the extension and let's add another one, later, after the GUID
                        string basePath = Path.Combine(Path.GetDirectoryName(FilePath),
                            Path.GetFileNameWithoutExtension(FilePath));

                        for (int i = 0; i < NumberOfSubFiles; i++)
                        {
                            Guid fileGuid = Guid.NewGuid();
                            guids.Add(fileGuid);

                            string thisFilePath = String.Format(SUBFILENAME_TEMPLATE, basePath, fileGuid);
                            newSubFiles.Add(thisFilePath);
                            XmlWriter writer = null;

                            try
                            {
                                writer = new XmlTextWriter(thisFilePath, Encoding.UTF8);

                                // Root element
                                writer.WriteStartElement(SUBFILEWORDS_ELEMENT, XML_NAMESPACE);
                                // How many to write?
                                int size = countPerFile + (i < countOfExtras ? 1 : 0);
                                // Write the range of this file
                                SaveWordsToElement(writer, nowAt, size);
                                writer.WriteEndElement();

                                nowAt += size;
                            }
                            finally
                            {
                                if (writer != null)
                                {
                                    writer.Close();
                                }
                            }
                        }
                    }

                    XmlWriter mainWriter = null;
                    try
                    {
                        mainWriter = new XmlTextWriter(filePath, Encoding.UTF8);
                        // Root element
                        mainWriter.WriteStartElement(DOCUMENT_ELEMENT, XML_NAMESPACE);

                        // The two parameters controlling string comparisons
                        mainWriter.WriteElementString(COMPAREOPTIONS_ELEMENT, XML_NAMESPACE,
                            CompareOptions.ToString());
                        mainWriter.WriteElementString(CULTURE_ELEMENT, XML_NAMESPACE, Culture.ToString());

                        mainWriter.WriteStartElement(WORDS_ELEMENT, XML_NAMESPACE);
                        // Write all the words, or write a list of sub-files. The format supports both at once, but we
                        // never mix them when saving.
                        if (NumberOfSubFiles == 0)
                        {
                            mainWriter.WriteStartElement(WORDSLIST_ELEMENT, XML_NAMESPACE);
                            SaveWordsToElement(mainWriter, 0, words.Count);
                            mainWriter.WriteEndElement();
                        }
                        else
                        {
                            // Write a list of files
                            mainWriter.WriteStartElement(FILES_ELEMENT, XML_NAMESPACE);

                            // Use relative paths
                            string fileName = Path.GetFileNameWithoutExtension(filePath);

                            foreach (Guid guid in guids)
                            {
                                mainWriter.WriteElementString(FILEPATH_ELEMENT, XML_NAMESPACE,
                                    String.Format(SUBFILENAME_TEMPLATE, fileName, guid));
                            }
                            // The two WriteEndElements could be factored out for the if branches, but they end
                            // different elements and so logically belong inside the clause.
                            mainWriter.WriteEndElement();
                        }
                        // </Document>
                        mainWriter.WriteEndElement();
                    }
                    finally
                    {
                        if (mainWriter != null)
                        {
                            mainWriter.Close();
                        }
                    }

                    // Delete the old files
                    foreach (string deleteMe in subFiles)
                    {
                        // Never fail due to inability to remove a previous file.
                        try
                        {
                            File.Delete(deleteMe);
                        }
                        catch (Exception)
                        {
                        }
                    }

                    // Set the new list of sub files
                    subFiles = newSubFiles;
                }
            }
            catch (Exception e)
            {
                throw new SaveIndexDataException(e, "Error saving.");
            }
        }

        /// <summary>
        /// This method is implemented to compare two strings using the culture info contained in this class. This
        /// method is used by the internal data store when sorting the words in the list. We also use the
        /// <see cref="CompareOptions"/> property here.
        /// </summary>
        /// <param name="x">One string to compare against <paramref name="y"/>.</param>
        /// <param name="y">One string to compare against <paramref name="x"/>.</param>
        /// <exception cref="ArgumentException">If any argument is not a string (null values are allowed).</exception>
        /// <returns>Negative if x &lt; y, positive if y &gt; x, zero if equal.</returns>
        public int Compare(object x, object y)
        {
            string xStr = x as string;
            string yStr = y as string;

            // Compare is called numerous times while handling the SortedList. The parameters are almost never
            // non-strings or null, and this is an ever so slight optimization for that fact. (It helps real run time
            // a tiny bit, at least on P6-based CPUs.)
            if (xStr == null || yStr == null)
            {
                if (xStr == null && x != null)
                {
                    throw new ArgumentException("Not a string.", "x");
                }
                if (yStr == null && y != null)
                {
                    throw new ArgumentException("Not a string.", "y");
                }
            }

            return culture.CompareInfo.Compare(xStr, yStr, compareOptions);
        }

        /// <summary>
        /// Gets/sets the <see cref="IList"/> of integer positions for a specific word string.
        /// </summary>
        /// <param name="word">The word to change/read the list for.</param>
        /// <exception cref="ArgumentException">If <paramref name="word"/> is empty.</exception>
        /// <exception cref="ArgumentNullException">If <paramref name="word"/> is null.</exception>
        /// <exception cref="ArgumentNullException">If <paramref name="value"/> is null (only setter).</exception>
        /// <exception cref="ArgumentException">If any element in <paramref name="value"/> is null/impossible to cast
        /// into a <see cref="long"/>.
        /// </exception>
        /// <value>
        /// An <see cref="IList"/> of <see cref="long"/> values.
        /// </value>
        public IList this[string word]
        {
            get
            {
                if (word == null)
                {
                    throw new ArgumentNullException("word");
                }
                word = word.Trim();
                if (word.Length == 0)
                {
                    throw new ArgumentException("Empty string specified.", "word");
                }

                IList original = (IList) words[word];
                if (original == null)
                {
                    return null;
                }

                return new ArrayList(original);
            }
            set
            {
                if (value == null)
                {
                    throw new ArgumentNullException("value");
                }
                if (word == null)
                {
                    throw new ArgumentNullException("word");
                }
                word = word.Trim();
                if (word.Length == 0)
                {
                    throw new ArgumentException("Empty string specified.", "word");
                }

                IList newList = new ArrayList(value.Count);
                try
                {
                    // We use Convert.ToInt64 instead of just casting, as it's easy to imagine the user entering values
                    // as simple ints, while the internal type here is long. Boxing means that (Int64) element if
                    // element is an Int32, would fail.
                    foreach (object element in value)
                    {
                        if (element == null)
                        {
                            throw new ArgumentException("Null values not allowed.", "value");
                        }
                        newList.Add(Convert.ToInt64(element));
                    }
                }
                catch (InvalidCastException e)
                {
                    throw new ArgumentException("Only long values allowed.", "value", e);
                }

                words[word] = newList;
            }
        }

        /// <summary>
        /// Private helper to load the entries from a specific subfile.
        /// </summary>
        /// <param name="filePath">The subfile to load from.</param>
        /// <exception cref="Exception">This is a private helper, parameters are assumed to be valid and exceptions are
        /// assumed to be wrapped by the caller.</exception>
        private void GetWordsFromFile(string filePath)
        {
            XmlReader reader = null;

            try
            {
                reader = new XmlTextReader(filePath);

                bool readRoot = false;
                while (reader.Read())
                {
                    if (reader.NodeType == XmlNodeType.Element)
                    {
                        // Document
                        reader.ReadStartElement(SUBFILEWORDS_ELEMENT, XML_NAMESPACE);

                        if (readRoot)
                        {
                            throw new XmlException(String.Format("Double root elements in {0}.", filePath));
                        }
                        readRoot = true;

                        GetWordsFromElement(reader);
                    }
                }
                if (!readRoot)
                {
                    throw new XmlException(String.Format("No root element found in {0}.", filePath));
                }
            }
            finally
            {
                if (reader != null)
                {
                    reader.Close();
                }
            }
        }

        /// <summary>
        /// Private helper to load the entries in a list intermixing like Word/Locations/Word/Locations...
        /// </summary>
        /// <param name="reader">The <see cref="XmlReader"/> we are reading from.</param>
        /// <exception cref="Exception">This is a private helper, parameters are assumed to be valid and exceptions are
        /// assumed to be wrapped by the caller.</exception>
        private void GetWordsFromElement(XmlReader reader)
        {
            do
            {
                // We've found a Word. The if clause will not end until </Locations> is reached.
                if (reader.NodeType == XmlNodeType.Element)
                {
                    string word = reader.ReadElementString(WORD_ELEMENT, XML_NAMESPACE).Trim();

                    ReadToStartElement(reader);
                    reader.ReadStartElement(LOCATIONS_ELEMENT, XML_NAMESPACE);

                    IList wordList = (IList) words[word];

                    // Not already present, prepare a new list
                    if (wordList == null)
                    {
                        wordList = new ArrayList();
                    }

                    do
                    {
                        if (reader.NodeType == XmlNodeType.Element)
                        {
                            // Int64 Parse will always trim on its own
                            wordList.Add(Int64.Parse(reader.ReadElementString(LOCATION_ELEMENT, XML_NAMESPACE)));
                        }
                        else
                        {
                            // Consume one token
                            reader.Read();
                        }
                    }
                    while (reader.NodeType != XmlNodeType.EndElement);

                    words[word] = wordList;
                }
            }
            while (reader.Read() && reader.NodeType != XmlNodeType.EndElement);
        }

        /// <summary>
        /// Saves a subset of the stored words, in sorted order, to the specified <see cref="XmlWriter"/>.
        /// </summary>
        /// <param name="writer">The writer instance being the target for our output.</param>
        /// <param name="from">The index of the first word to write.</param>
        /// <param name="count">The count of words to write.</param>
        /// <exception cref="Exception">This is a private helper, parameters are assumed to be valid and exceptions are
        /// assumed to be wrapped by the caller.</exception>
        private void SaveWordsToElement(XmlWriter writer, int from, int count)
        {
            // Write a Word, a Locations element containing individual Location values, for each word to be written.
            for (int i = from; i < from + count; i++)
            {
                writer.WriteElementString(WORD_ELEMENT, XML_NAMESPACE, (string) words.GetKey(i));

                writer.WriteStartElement(LOCATIONS_ELEMENT, XML_NAMESPACE);

                IDictionary avoidDuplicates = new Hashtable();
                foreach (long value in (IList) words.GetByIndex(i))
                {
                    // Only add once, even if many would exist
                    if (!avoidDuplicates.Contains(value))
                    {
                        avoidDuplicates.Add(value, null);
                        writer.WriteElementString(LOCATION_ELEMENT, XML_NAMESPACE, value.ToString());
                    }
                }
                writer.WriteEndElement();
            }
        }
    }
}
