/*
 * Copyright (c) 2006, TopCoder, Inc. All rights reserved.
 *
 * Author: TCSDEVELOPER
 * Version: 1.0
 */

using NUnit.Framework;
using System;
using System.Collections;
using System.Globalization;
using TopCoder.Document.Index.WordStreams;

namespace TopCoder.Document.Index.Persistence.DB
{
    /// <summary>
    /// This class tests <see cref="DBCollectionIndexSource"/>.
    /// </summary>
    /// <remarks>
    /// This class "implements" <see cref="IParameters"/> to supply itself as an (incorrect) parameter object.
    /// </remarks>
    [TestFixture]
    public class TestDBCollectionIndexSource : DBTests, IParameters
    {
        /// <summary>
        /// A test object used in several tests, created based on a real document table.
        /// </summary>
        private DBCollectionIndexSource testObject1;

        /// <summary>
        /// A test object crated with no loaded table.
        /// </summary>
        private DBCollectionIndexSource testObject2;

        /// <summary>
        /// An <see cref="IndexInfo"/> used for testing.
        /// </summary>
        private IndexInfo index;

        /// <summary>
        /// Sets up the testing fixture.
        /// </summary>
        [SetUp]
        public override void Setup()
        {
            base.Setup();

            index = new IndexInfo("TopCoder.Document.Index.Persistence.Test",
                "TopCoder.Document.Index.Persistence.DB.DBDocumentIndexSource",
                new DBDocumentIndexParameters("Test", 9));
            testObject1 = new DBCollectionIndexSource();
            testObject1.Load(new DBPersistenceParameters("Test"));

            testObject2 = new DBCollectionIndexSource();
        }

        /// <summary>
        /// Counts the elements reported by an enumerator.
        /// </summary>
        /// <param name="enumerator">The enumerator.</param>
        /// <returns>The number of elements.</returns>
        private static int CountEnumerator(IEnumerator enumerator)
        {
            int count = 0;
            while (enumerator.MoveNext())
            {
                count++;
            }

            return count;
        }

        /// <summary>
        /// Tests constructing with a null namespace.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestNameSpaceConstructorNull()
        {
            new DBCollectionIndexSource(null);
        }

        /// <summary>
        /// Tests constructing with an empty namespace.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestNameSpaceConstructorEmpty()
        {
            new DBCollectionIndexSource(" ");
        }

        /// <summary>
        /// Tests constructing with a non-existing namespace.
        /// </summary>
        [Test]
        [ExpectedException(typeof(DBPersistenceConfigurationException))]
        public void TestNameSpaceConstructorInvalid()
        {
            new DBCollectionIndexSource("Some.Weird.Namespace");
        }

        /// <summary>
        /// Tests constructing with an existing namespace, but not the required parameters.
        /// </summary>
        [Test]
        [ExpectedException(typeof(DBPersistenceConfigurationException))]
        public void TestNameSpaceConstructorMissingParam()
        {
            new DBCollectionIndexSource("TopCoder.Document.Index.Persistence.NoID");
        }

        /// <summary>
        /// Tests the namespace constructor, we verify that an exception is thrown when trying to Add data, because
        /// the connection name is invalid.
        /// </summary>
        [Test]
        [ExpectedException(typeof(SaveIndexDataException))]
        public void TestNameSpaceConstructor()
        {
            DBCollectionIndexSource source = new DBCollectionIndexSource("TopCoder.Document.Index.Persistence.DB.2");
            source.Add("document", new IndexInfo(typeof(DBDocumentIndexSource).Assembly.FullName,
                typeof(DBDocumentIndexSource).FullName, new DBDocumentIndexParameters("AnotherConnection", 4)));
        }

        /// <summary>
        /// Tests adding indices.
        /// </summary>
        [Test]
        public void TestAdd()
        {
            testObject1.Add("newindex", index);
            Assert.AreEqual(index, testObject1["newindex"], "The stored index should be returned.");
            testObject1.Remove("newindex");
            testObject1.Save();
            // Commit the transaction, so another conneciton can modify the same data, remove to avoid document ID
            // conflicts

            testObject2.Add("index", index);
            Assert.AreEqual(index, testObject2["index"], "The stored index should be returned.");
        }

        /// <summary>
        /// The index info can store a reference to just about any class. However, we can only faithfully reproduce
        /// references to <see cref="DBDocumentIndexSource"/>. Here, we reference an assembly which doesn't exist.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestAddInvalidAssembly()
        {
            IndexInfo index = new IndexInfo("TopCoder.Document.IDontExist.Persistence.Test",
                "TopCoder.Document.Index.Persistence.DB.DBDocumentIndexSource",
                new DBPersistenceParameters("Test"));
            testObject1.Add("newindex", index);
        }

        /// <summary>
        /// The index info can store a reference to just about any class. However, we can only faithfully reproduce
        /// references to <see cref="DBDocumentIndexSource"/>. Here, we reference the wrong class.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestAddWrongClass()
        {
            IndexInfo index = new IndexInfo("TopCoder.Document.Index.Persistence.Test",
                "TopCoder.Document.Index.Persistence.DB.DBCollectionIndexSource",
                new DBPersistenceParameters("Test"));
            testObject1.Add("newindex", index);
        }

        /// <summary>
        /// Tests adding a null index. Exception expected.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestAddNull1()
        {
            testObject1.Add("newname", null);
        }

        /// <summary>
        /// Tests adding a null index name. Exception expected.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestAddNull2()
        {
            testObject1.Add(null, new IndexInfo("TopCoder.Document.Index.Persistence.Test",
                "TopCoder.Document.Index.Persistence.DB.DBDocumentIndexSource",
                new DBPersistenceParameters("Test")));
        }

        /// <summary>
        /// Tests adding an empty index name. Exception expected.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestAddEmpty()
        {
            testObject1.Add(" ", new IndexInfo("TopCoder.Document.Index.Persistence.Test",
                "TopCoder.Document.Index.Persistence.DB.DBDocumentIndexSource",
                new DBPersistenceParameters("Test")));
        }

        /// <summary>
        /// Tests removing an index, both one which really exists and one which does not.
        /// </summary>
        [Test]
        public void TestRemove()
        {
            testObject1.Remove("index");
            Assert.IsNull(testObject1["index"], "Index 'index' should now be null.");

            Assert.AreEqual(1, CountEnumerator(testObject1.GetEnumerator()), "Index count not as expected. 1");

            // The count should remain zero, but no exception shold be thrown
            testObject2.Remove("entre");
            Assert.IsNull(testObject2["entre"], "Index 'index' should still be null.");
            Assert.AreEqual(0, CountEnumerator(testObject2.GetEnumerator()), "Index count not as expected. 2");
        }

        /// <summary>
        /// Tests removing a null index name. This should raise an exception.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestRemoveNull()
        {
            testObject1.Remove(null);
        }

        /// <summary>
        /// Tests removing an empty name. This should raise an exception.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestRemoveEmpty()
        {
            testObject1.Remove(" ");
        }

        /// <summary>
        /// A simple test of <see cref="DBCollectionIndexSource.Contains"/>.
        /// </summary>
        [Test]
        public void TestContains()
        {
            Assert.IsTrue(testObject1.Contains("index"), "Index 'index' should exist.");
            Assert.IsFalse(testObject2.Contains("index"), "Index 'index' should not exist.");
        }

        /// <summary>
        /// Tests checking for a null index name. This should raise an exception.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestContainsNull()
        {
            testObject1.Contains(null);
        }

        /// <summary>
        /// Tests checking for an empty index name. This should raise an exception.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestContainsEmpty()
        {
            testObject1.Contains(" ");
        }

        /// <summary>
        /// Tests clearing the store.
        /// </summary>
        [Test]
        public void TestClear()
        {
            testObject1.Clear();
            Assert.IsNull(testObject1["index"], "Known index should now be null.");
            Assert.AreEqual(0, CountEnumerator(testObject1.GetEnumerator()), "Count should be zero after clearing.");
        }

        /// <summary>
        /// Tests <see cref="DBCollectionIndexSource.GetEnumerator"/>, by iterating in order.
        /// </summary>
        [Test]
        public void TestGetEnumerator()
        {
            IDictionaryEnumerator enumerator = testObject1.GetEnumerator();
            for (int i = 0; i < 2; i++)
            {
                // Now, the specific order here WILL be dependent on the hashtable
                enumerator.MoveNext();
                Assert.AreEqual("index2", enumerator.Key, "The first key doesn't match.");
                Assert.AreEqual("Test",
                    (((IndexInfo) enumerator.Value).Parameters as DBDocumentIndexParameters).ConnectionName,
                    "Wrong connection name for first index.");
                Assert.AreEqual(2,
                    (((IndexInfo) enumerator.Value).Parameters as DBDocumentIndexParameters).DocumentID,
                    "Wrong document ID for first index.");

                enumerator.MoveNext();
                Assert.AreEqual("index", enumerator.Key, "The second key doesn't match.");
                Assert.AreEqual("Test",
                    (((IndexInfo) enumerator.Value).Parameters as DBDocumentIndexParameters).ConnectionName,
                    "Wrong connection name for second index.");
                Assert.AreEqual(1,
                    (((IndexInfo) enumerator.Value).Parameters as DBDocumentIndexParameters).DocumentID,
                    "Wrong document ID for second index.");

                Assert.IsFalse(enumerator.MoveNext(), "Moving next should fail after end.");
                enumerator.Reset();
            }
        }

        /// <summary>
        /// Tests loading new data, after first adding an index.
        /// </summary>
        [Test]
        public void TestLoad()
        {
            testObject1.Dispose();

            testObject2.Add("newindex", index);
            testObject2.Load(new DBPersistenceParameters("Test"));

            testObject1 = testObject2;
            // Will verify that testObject1 now matches testObject2
            TestGetEnumerator();
        }

        /// <summary>
        /// Tests loading with a null reference. Exception is thrown.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestLoadNull()
        {
            testObject1.Load(null);
        }

        /// <summary>
        /// Tests loading with a parameter object of incorrect type.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestLoadWrongType()
        {
            // this implements IParameters
            testObject1.Load(this);
        }

        /// <summary>
        /// It should be possible to save the default object.
        /// </summary>
        [Test]
        public void TestSaveDefault()
        {
            testObject2.Save();
        }

        /// <summary>
        /// Add an index, commit, dispose, load in a new object and verify the results.
        /// </summary>
        [Test]
        public void TestSave()
        {
            testObject1.Add("newdoc", index);

            // Save to test, Dispose for general resource management
            testObject1.Save();
            testObject1.Dispose();

            testObject1 = new DBCollectionIndexSource();
            // Well, a DBDocumentIndexParameters instance is also OK, in addition to DBPersistenceParameters
            testObject1.Load(new DBDocumentIndexParameters("Test", 1));

            IndexInfo info = testObject1["newdoc"];
            DBDocumentIndexParameters parameters = info.Parameters as DBDocumentIndexParameters;

            Assert.AreEqual("Test", parameters.ConnectionName, "Connection name should match.");
            Assert.AreEqual(9, parameters.DocumentID, "Document ID should match.");
        }

        /// <summary>
        /// Add an index, dispose, load in a new object and verify that nothing has changed. (Dispose will roll-back.)
        /// </summary>
        [Test]
        public void TestDispose()
        {
            testObject1.Add("newdoc", index);

            testObject1.Dispose();

            testObject1 = new DBCollectionIndexSource();
            // Well, a DBDocumentIndexParameters instance is also OK, in addition to DBPersistenceParameters
            testObject1.Load(new DBDocumentIndexParameters("Test", 1));

            IndexInfo info = testObject1["newdoc"];
            Assert.IsNull(info, "newdoc should be null.");
        }

        /// <summary>
        /// The getter part of the indexer is quite simple, we just verify that we retrieve correct data.
        /// </summary>
        [Test]
        public void TestGetIndexer()
        {
            IndexInfo index = testObject1["index"];
            Assert.AreEqual("Test",
                    (index.Parameters as DBPersistenceParameters).ConnectionName,
                    "Wrong connection name.");

            Assert.IsNull(testObject1["I DON'T EXIST"],
                "Null should be returned for those indices that are not present.");
        }

        /// <summary>
        /// Tests getting the list for a null index name.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestGetIndexerNull()
        {
            IndexInfo list = testObject1[null];
        }

        /// <summary>
        /// Tests getting the list for an empty index name.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestGetIndexerEmpty()
        {
            IndexInfo list = testObject1[" "];
        }

        /// <summary>
        /// Tests setting index infos with the indexer.
        /// </summary>
        [Test]
        public void TestSetIndexer()
        {
            testObject1["index"] = index;
            Assert.AreEqual(index, testObject1["index"], "The stored index should be returned.");
            testObject1.Remove("index");
            testObject1.Save();
            // Remove to be able to do another operation

            testObject2["index"] = index;
            Assert.AreEqual(index, testObject2["index"], "The stored index should be returned.");
        }

        /// <summary>
        /// Tests setting the list for a null index name. An exception is thrown.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestSetIndexerNullIndexName()
        {
            testObject1[null] = index;
        }

        /// <summary>
        /// Tests setting the list for an empty index name. An exception is thrown.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentException))]
        public void TestSetIndexerEmptyIndexName()
        {
            testObject1[" "] = index;
        }

        /// <summary>
        /// Tests setting a null index info. An exception is thrown.
        /// </summary>
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void TestSetIndexerNullList()
        {
            testObject1["someindex"] = null;
        }

        /// <summary>
        /// Closes active connections. This was not specified by the design, but it is dearly needed if the connection
        /// limit on the Informix server is low.
        /// </summary>
        [TearDown]
        public override void Teardown()
        {
            testObject1.Dispose();
            testObject2.Dispose();
            base.Teardown();
        }
    }
}
