/*
 * Copyright (c) 2006, TopCoder, Inc. All rights reserved.
 *
 * Author: TCSDEVELOPER, Ghostar
 * Version: 1.0
 */

using NUnit.Framework;
using System;
using System.Collections;
using System.IO;
using TopCoder.Document.Index.Persistence.DB;
using TopCoder.Document.Index.Persistence.Xml;
using TopCoder.Document.Index.WordStreams;

namespace TopCoder.Document.Index.Persistence
{
    /// <summary>
    /// Demo tests cases, two very similar cases, one for XML and one for database access.
    ///
    /// These tests use test_files\config.xml for basic configuration of the Connection Factory in the database tests.
    /// Please see the documentation for that component and the component spec for more information about that file.
    ///
    /// The <c>DBTests</c> base class creates a simple test database configuration. It also cleans up. The correct
    /// database schema must be present, and the connection string set in config.xml.
    /// </summary>
    [TestFixture]
    public class Demo : DBTests
    {
        /// <summary>
        /// Tests XML access, a series of simple and common operations.
        /// </summary>
        [Test]
        public void XMLAccess()
        {
            // Load from an XML file
            ICollectionIndexSource xmlSource = new XmlCollectionIndexSource();
            IParameters parameters = new SimpleIndexParameters(@"..\..\test_files\testcollections.xml");
            xmlSource.Load(parameters);

            // Enumerate through the indices
            IDictionaryEnumerator enumerator = xmlSource.GetEnumerator();

            while (enumerator.MoveNext())
            {
                Console.WriteLine("Key: {0}, Filename: {1}", enumerator.Key,
                    ((enumerator.Value as IndexInfo).Parameters as SimpleIndexParameters).FileName);
            }

            // Add a new index
            IParameters documentParams = new SimpleIndexParameters(@"..\..\test_files\testarchive.xml");
            IndexInfo info = new IndexInfo("TopCoder.Document.Index.Persistence.Test",
                "TopCoder.Document.Index.Persistence.Xml.XmlDocumentIndexSource", documentParams);

            xmlSource.Add("Document", info);

            // Save the changes, not done here to preserve the demo files intact
            // xmlSource.Save();

            //Remove the index
            xmlSource.Remove("Document");

            // Get an index
            IndexInfo retrieved = xmlSource["Document2"];

            // Clear all indices
            xmlSource.Clear();

            // Save the changes, not done here to preserve the demo files intact
            // xmlSource.Save();

            // Create a new XmlDocumentIndexSource
            XmlDocumentIndexSource source = new XmlDocumentIndexSource();
            source.Load(documentParams);

            // Add a new word
            WordInfo word = new WordInfo("Cat", 102);
            source.Add(word);

            // Remove a word
            source.Remove("Dog");

            // Save the changes, not done here to preserve the demo files intact
            // source.Save();

            // Save to multiple files
            source.NumberOfSubFiles = 2;
            // source.Save();

            // Check if the document contains a word
            if (source.Contains("Snake"))
            {
                Console.WriteLine("Snake exists, eeeek!");
            }

            // Clear all words
            source.Clear();

            try
            {
                if (!Directory.Exists(@"..\..\test_files\savedir"))
                {
                    Directory.CreateDirectory(@"..\..\test_files\savedir");
                }

                //Set a new file path
                source.FilePath = @"..\..\test_files\savedir\Empty.xml";

                //Save the empty file
                source.Save();
            }
            finally
            {
                File.Delete(source.FilePath);
            }
        }

        /// <summary>
        /// Tests database access, for an index collection.
        /// </summary>
        [Test]
        public void DBAccessCollection()
        {
            // Load from a database
            using (DBCollectionIndexSource dbSource = new DBCollectionIndexSource())
            {
                // Test is a defined connection name
                IParameters parameters = new DBPersistenceParameters("Test");

                dbSource.Load(parameters);

                // Enumerate through the indices
                IDictionaryEnumerator enumerator = dbSource.GetEnumerator();

                while (enumerator.MoveNext())
                {
                    Console.WriteLine("Key: {0}, document ID: {1}", enumerator.Key,
                                        ((enumerator.Value as IndexInfo).Parameters as DBDocumentIndexParameters).
                                        DocumentID);
                }

                // Add a new index
                IParameters documentParams = new DBDocumentIndexParameters("Test", 5);
                IndexInfo info = new IndexInfo("TopCoder.Document.Index.Persistence.Test",
                    "TopCoder.Document.Index.Persistence.DB.DBDocumentIndexSource", documentParams);

                dbSource.Add("Document", info);

                // Save the changes
                dbSource.Save();

                // Remove the index
                dbSource.Remove("Document");

                // Get an index
                IndexInfo retrieved = dbSource["Document2"];

                // Clear all indices
                dbSource.Clear();

                // NOTE: If we don't save, any changes are lost.
                dbSource.Save();
            }
        }

        /// <summary>
        /// Access to an individual document.
        /// </summary>
        [Test]
        public void DBAccessIndividualDocument()
        {
            //Create a new DBDocumentIndexSource
            using (DBDocumentIndexSource source = new DBDocumentIndexSource())
            {
                IParameters documentParams = new DBDocumentIndexParameters("Test", 1);
                source.Load(documentParams);

                // Add a new word
                WordInfo word = new WordInfo("Cat", 102);
                source.Add(word);

                // Remove a word
                source.Remove("Dog");

                // Save the changes
                source.Save();

                // Get a specific words locations
                IList locations = source["Horse"];

                // Check if the document contains a word
                if (source.Contains("Snake"))
                {
                    Console.WriteLine("Snake exists, eeeek!");
                }

                // Clear all words
                source.Clear();
            }
        }
    }
}
